"""This module provides the `TWDA` singleton: Tournament Wining Decks Archive.

If it has not been initialized, TWDA will evaluate to False.
TWDA must be configured with `TWDA.configure()` before being used.
"""
import collections
import io
import itertools
import logging
import os
import pickle
import re

import arrow
import arrow.parser
import requests

from . import config
from . import deck
from . import vtes

logger = logging.getLogger()


class AnalysisError(Exception):
    pass


# Headers used in TWDA, e.g. "Action modifiers/Combat cards total: (4)"
HEADERS = set()
HEADERS.update(config.HEADERS)
HEADERS.update([h + "s" for h in config.HEADERS] + ["ally", "allies"])
HEADERS.update(
    ["{}/{}".format(h1, h2) for h1, h2 in itertools.permutations(HEADERS, 2)]
)
HEADERS.update(config.ADDITIONAL_HEADERS)
HEADERS.update([h + " card" for h in HEADERS] + [h + " cards" for h in HEADERS])
HEADERS.update(["card", "cards", "deck"])
HEADERS.update([h + " total" for h in HEADERS])
HEADERS.update([h + ":" for h in HEADERS])


class _TWDA(collections.OrderedDict):
    """ An OrderedDict of the TWDA. Parsing TWDA.html is the hard part.

    Attributes:
        spoilers (dict): cards played in over 25% of decks
        tail_re (str): regexp used to parse tail part of card line
    """

    def load_from_vekn(self, limit=None, save=True):
        """Load from vekn.net

        Args:
            limit: Maximum number of decks to load (used to speed up tests)
        """
        r = requests.request("GET", config.VEKN_TWDA_URL)
        r.raise_for_status()
        self.load_html(io.StringIO(r.content.decode("utf-8")), limit, save)

    def load_html(self, source, limit=None, save=True):
        """Load from TWDA.html

        The TWDA is then pickled for future use, to avoid loading it too often.

        Args:
            source (file): The HTML
            limit: Maximum number of decks to load (used to speed up tests)
        """
        self.clear()
        # used to match a card even if a comment, discipline or clan has been added.
        self.tail_re = r"(?P<card_name>.+?)\s+(((\(|\[|-|/)(?P<comment>.+))|(({})\s*$))"
        self.tail_re = self.tail_re.format(
            "|".join(
                re.escape(s.lower())
                for s in vtes.VTES.clans + vtes.VTES.disciplines + ["trifle"]
            )
        )
        for count, (lines, twda_id) in enumerate(self._get_decks_html(source), 1):
            try:
                self[twda_id] = self._load_deck_html(lines, twda_id)
            except arrow.parser.ParserError as e:
                logger.error(e)
            if limit and count >= limit:
                break
        logger.info("TWDA loaded")
        if save:
            pickle.dump(TWDA, open(config.TWDA_FILE, "wb"))

    def configure(self, date_from=None, date_to=None, min_players=0, spoilers=True):
        """Prepare the TWDA, taking date and players count filters into account.

        Also compute `self.spoilers`.

        Args:
            date_from (datetime): filter out decks before this date
            date_to (datetime): filter out decks after this date
            min_players (int): if > 0, filter out decks with less players
            spoilers (bool): if True, compute spoilers to filter them out for analysis
        """
        if date_from:
            for key in [key for key, value in self.items() if value.date < date_from]:
                del self[key]
        if date_to:
            for key in [key for key, value in self.items() if value.date >= date_to]:
                del self[key]
        if min_players:
            for key in [
                key
                for key, value in self.items()
                if value.players_count and value.players_count < min_players
            ]:
                del self[key]
        if spoilers and len(self) > 50:
            self.spoilers = {
                name: count / len(self)
                for name, count in collections.Counter(
                    itertools.chain.from_iterable(d.keys() for d in self.values())
                ).items()
                if count > len(self) / 4
            }
            logger.debug("Spoilers: {}".format(self.spoilers))
        else:
            self.spoilers = {}

    @staticmethod
    def _get_decks_html(source):
        """Get lines for each deck, using HTML tags as separators.

        Args:
            source (file): HTML source file-like object implementing `readlines()`
        """
        lines = []
        for line_num, line in enumerate(source.readlines(), start=1):
            try:
                twda_id = re.match(r"^<a id=([^\s]*)\s", line).group(1)
            except AttributeError:
                pass
            if re.match(r"^<hr><pre>$", line):
                lines = []
            elif re.match(r"^</pre>", line):
                yield lines, twda_id
            else:
                lines.append((line_num, line[:-1]))

    def _load_deck_html(self, lines, twda_id):
        """Parse TWDA.html lines to build a deck.

        Args:
            lines (list): lines of the deck as a list of `str`
            twda_id (str): the TWDA ID for this deck
        """
        current = deck.Deck()
        # First three lines are stable throughout TWDA
        current.event = lines[0][1]
        current.place = lines[1][1]
        current.date = arrow.get(lines[2][1], "MMMM Do YYYY")
        start = 3
        # tournament format and players count are not always set
        try:
            regexp = r"^(\d+R\+F)"
            current.tournament_format = re.match(regexp, lines[start][1]).group(1)
            start += 1
        except AttributeError:
            pass
        try:
            players_count = re.match(r"^(\d+|\?+)\s*player", lines[start][1]).group(1)
            start += 1
            current.players_count = int(players_count)
        except (AttributeError, ValueError):
            pass
        # Player is always set, but after tournament format and count if any
        current.player = lines[start][1]
        start += 1
        # Ignore Organizer line (rare inclusion)
        try:
            if re.match(r"^\s*(O|o)rgani(s|z)er", lines[start][1]):
                start += 1
        except AttributeError:
            pass
        # Newer lists provide an event link
        try:
            current.event_link = re.match(r"^(https?://.*)$", lines[start][1]).group(1)
            start += 1
        except AttributeError:
            pass
        # After this stable header, comments and format tricks happen
        for line_num, original_line in lines[start + 1 :]:
            # remove misplaced spaces around punctuation for easy parsing
            line = original_line.strip()
            line = line.replace(" :", ":")
            line = line.replace("( ", "(")
            line = line.replace(" )", ")")
            line = line.replace(" /", "/")
            line = line.replace("/ ", "/")
            if not line:
                current.comment_end()
                continue
            # separtor lines ("---" or "===") are used to detect the beginning of the
            # actual deck list: this is the most reliable method as the "Crypt" word
            # can appear in deck comments.
            if line and len(line) > 2 and set(line).issubset({"=", "-"}):
                current.comment_end()
                current.separator = True
                continue
            # if we did not hit any separator (e.g. "=============="),
            # try to find deck name, author and comments
            if not current.separator:
                if not current.score:
                    try:
                        score = re.match(
                            r"-?-?\s*((?P<round_wins>\d)\s*(G|g)(W|w)\s*"
                            r"(?P<round_vps>\d(\.|,)?\d?)\s*((v|V)(p|P))?\s*\+?\s*)?"
                            r"(?P<final>\d(\.|,)?\d?)\s*(v|V)(p|P)",
                            line,
                        )
                        if score.group("round_wins"):
                            current.score = "{}gw{} + {}vp in the final".format(
                                score.group("round_wins"),
                                score.group("round_vps"),
                                score.group("final"),
                            )
                        else:
                            current.score = "{}vp in the final".format(
                                score.group("final")
                            )
                        continue
                    except AttributeError:
                        pass
                if not current.name:
                    try:
                        current.name = re.match(
                            r"^\s*(?:(?:d|D)eck)?\s?(?:n|N)ame\s*:\s*(.*)$", line
                        ).group(1)
                        continue
                    except AttributeError:
                        pass
                if not current.author:
                    try:
                        current.author = re.match(
                            r"(?:(?:(?:(?:c|C)reated)|(?:(?:d|D)eck))"
                            r"\s*(?:b|B)y|(?:a|A)uthors?|(?:c|C)reators?)\s*:?\s*(.*)$",
                            line,
                        ).group(1)
                        continue
                    except AttributeError:
                        pass
                if not line:
                    continue
                # Do not consider the "Crypt" line as part of the comments
                if re.match(r"^(c|C)rypt", line):
                    continue
                # Anything else is considered a comment until a separator is hit
                current.maybe_comment_line(line_num, original_line)
                continue

            # comments detection
            # wait for card identification for all trailing comments
            comment = ""
            multiline = False
            if "/*" in line:
                line, comment = line.split("/*", 1)
                multiline = True
            if "*/" in comment:
                comment, tail = comment.split("*/", 1)
                line = line + tail
                multiline = False
            if "*/" in line:
                comment_tail, line = line.split("*/", 1)
                current.maybe_comment_line(line_num, comment_tail)
                current.comment_end()
            if line and line[0] in "([/" and line[-1] in "/])":
                current.comment_begin(line_num, " " + line[1:-1].strip(), marker=True)
                line = ""
            # inline comments "--" or "//"
            # special case for "Bang Nakh -- Tiger's Claws"
            comment_re = (
                r"^(?P<line>.*?)\s*"
                r"(--|//)\s*"
                r"(?!(T|t)iger'?s? (C|c)laws?)"
                r"(?P<comment>.*?)\s*$"
            )
            match = re.match(comment_re, line)
            if match:
                line = match.group("line")
                comment = match.group("comment")
            if not line:
                if comment:
                    current.comment_begin(
                        line_num, comment, marker=True, multiline=multiline
                    )
                continue
            # lower all chars for easier parsing
            name, count, explicit = _get_card(line.lower())
            if name and count:
                if name in HEADERS:
                    # discard comment on header line (most likely card count)
                    continue
                if not explicit and name in set(
                    a.lower() for a in vtes.VTES.clans + vtes.VTES.disciplines
                ):
                    logger.warning(f"[{line_num}] improper discipline [{line}]")
                    continue
                card = None
                try:
                    card = vtes.VTES[name]
                except KeyError:
                    match = re.match(self.tail_re, name)
                    if match:
                        name = match.group("card_name")
                        tail = match.group("comment")
                        if comment and tail:
                            comment += " "
                        if tail:
                            comment += tail.rstrip(" )]-/\\")
                    try:
                        card = vtes.VTES[name]
                        logger.info(
                            f"[{line_num:<6}] fuzzy [{card['Name']}] - [{line}]"
                        )
                    except KeyError:
                        pass
                finally:
                    if card:
                        name = vtes.VTES.get_name(card)
                        current.update({name: count})
                        if comment:
                            current.comment_begin(
                                line_num,
                                comment,
                                name,
                                marker=True,
                                multiline=multiline,
                            )
                        else:
                            # if no blank line, comments on following lines are
                            # attached to his card
                            current.comment_begin(line_num, "", name)
                    else:
                        current.maybe_comment_line(line_num, line)
            # should not happen: by default the whole line is captured by _get_card()
            else:
                current.maybe_comment_line(line_num, line)

        current.comment_end()
        # check deck composition rules
        library_count = current.cards_count(vtes.VTES.is_library)
        if library_count < 60:
            logger.warning(
                "[{:<6}] Deck #{} is missing library cards [{}]".format(
                    line_num, twda_id, current
                )
            )
        if library_count > 90:
            logger.warning(
                "[{:<6}] Deck #{} has too many cards [{}]".format(
                    line_num, twda_id, current
                )
            )
        if current.cards_count(vtes.VTES.is_crypt) < 12:
            logger.warning(
                "[{:<6}] Deck #{} is missing crypt cards [{}]".format(
                    line_num, twda_id, current
                )
            )
        return current

    def no_spoil(self, card):
        return not getattr(self, "spoilers", None) or card not in self.spoilers


def _get_card(line):
    """Find card name and count in a TWDA.html line

    There are a lot of circumvoluted cases to take into account, cf. tests.
    Capturing the card comments here has proven difficult,
    since comments take so many forms in TWDA.
    We just capture the whole line and try and extract the card number if possible.

    Args:
        line (str): Test line
    Returns:
        name (str): name of the card
        count (int): number of exemplaries
        explicit (bool): if False, count was not explicit
    """
    card_match = re.match(
        # beginning of line, optional punctuation
        r"^\s*(-|_)?((x|X|\*|_)?\s*"
        # count
        # 2 digits maximum to avoid miscounting "419 Operation" and such
        # negative lookahead to avoid matching part of a card name (2nd, 3rd, 4th, ...)
        r"(?P<count>\d{1,2})(?!((st)|(nd)|(rd)|(th))))?"
        # optional punctuation
        r"\s*((x|X|-)\s)?\*?\s*"
        # non-greedy qualifier for card name, matches what the tail expression does not
        # special case for "channel 10" to avoid parsing 10x "channel".
        # "local 1111" and such are OK: we only consider 1 or 2 digits as valid count.
        r"(?P<name>((channel 10)|.+?))"
        # open parentheses for optional tail expression (separated for clarity)
        r"(("
        # mandatory punctuation (beware of "AK-47", "Kpist m/45", ...)
        r"((\s|\(|\[|\s/|:)+|\s=+)(?P<post_count_symbol>-|x|\*|\s)*"
        # sometimes (old deck lists) count is after the card name
        # for vampires, this is the capacity of the vampire
        r"(?P<count_or_capacity>\d{1,2})"
        # negative lookahead to avoid matching part of a card name (2nd, 3rd, 4th, ...)
        # also ignore blood ("b") / pool ("p") cost sometimes indicated after card name
        # also special exception for "Pier 13, Port of Baltimore"
        r"(?!((st)|(nd)|(rd)|(th)|(.?\d)|b|p|(..?port)))"
        # closing parentheses for tail expression (separated for clarity)
        r")|\s*$)",
        line,
    )
    if not card_match:
        return None, None, False
    name = card_match.group("name").strip()
    # when count is not given on the line, default to 1 (common in old deck lists)
    count = int(card_match.group("count") or 0)
    explicit = True
    if not count:
        count = int(card_match.group("count_or_capacity") or 1)
        if not card_match.group("post_count_symbol"):
            explicit = False
    return name, count, explicit


try:
    if not os.path.exists(config.TWDA_FILE) or os.stat(config.TWDA_FILE).st_size == 0:
        raise FileNotFoundError(config.TWDA_FILE)
    TWDA = pickle.load(open(config.TWDA_FILE, "rb"))
except (FileNotFoundError, EOFError):
    TWDA = _TWDA()  # evaluates to False as it is empty



import json
import logging
import os
import urllib.parse

import arrow
import flask
import pkg_resources  # part of setuptools
import requests

from . import analyzer
from . import config
from . import twda
from . import vtes


class KRCG(flask.Flask):
    def make_default_options_response(self,):
        response = super().make_default_options_response()
        response.headers.add("Access-Control-Allow-Headers", "*")
        response.headers.add("Access-Control-Allow-Methods", "*")
        return response

    def process_response(self, response):
        response.headers.add("Access-Control-Allow-Origin", "*")
        return response


logger = logging.getLogger()
base = flask.Blueprint("base", "krcg")


def create_app(test=False):
    logger.addHandler(logging.StreamHandler())
    logger.setLevel(logging.WARNING)
    if not test:
        vtes.VTES.load_from_vekn(save=False)
        vtes.VTES.configure()
        logger.info("loading TWDA")
        twda.TWDA.load_from_vekn(save=False)
        twda.TWDA.configure()
    logger.info("launching app")
    app = KRCG(__name__)
    app.register_blueprint(base)
    return app


@base.route("/")
@base.route("/index.html")
def swagger():
    return flask.render_template("index.html")


@base.route("/openapi.yaml")
def openapi():
    return flask.render_template(
        "openapi.yaml", version=pkg_resources.require("krcg")[0].version,
    )


@base.route("/card/<text>")
def card(text):
    try:
        text = int(text)
    except ValueError:
        pass
    try:
        return flask.jsonify(vtes.VTES.normalized(vtes.VTES[text]))
    except KeyError:
        return "Card not found", 404


@base.route("/deck", methods=["POST"])
def deck_by_cards():
    data = flask.request.get_json() or {}
    twda.TWDA.configure(
        arrow.get(data.get("date_from") or "2008-01-01"),
        arrow.get(data.get("date_to") or None),
        data.get("players") or 0,
        spoilers=False,
    )
    decks = twda.TWDA
    if data and data.get("cards"):
        A = analyzer.Analyzer()
        try:
            A.refresh(
                *[vtes.VTES.get_name(vtes.VTES[card]) for card in data["cards"]],
                similarity=1,
            )
            decks = A.examples
        except analyzer.AnalysisError:
            return "No result in TWDA", 404
        except KeyError:
            return "Invalid card name", 400
    return flask.jsonify([vtes.VTES.deck_to_dict(v, k) for k, v in decks.items()])


@base.route("/deck/<twda_id>")
def deck_by_id(twda_id):
    if not twda_id:
        return "Bad Request", 400
    if twda_id not in twda.TWDA:
        return "Not Found", 404
    return flask.jsonify(vtes.VTES.deck_to_dict(twda.TWDA[twda_id], twda_id))


@base.route("/complete/<text>")
def complete(text):
    return flask.jsonify(vtes.VTES.complete(text))


@base.route("/card", methods=["POST"])
def card_search():
    data = flask.request.get_json() or {}
    result = set(card["Id"] for card in vtes.VTES.original_cards.values())
    for type_ in data.get("type") or []:
        result &= vtes.VTES.search["type"].get(type_.lower(), set())
    for clan in data.get("clan") or []:
        clan = config.CLANS_AKA.get(clan.lower()) or clan
        result &= vtes.VTES.search["clan"].get(clan.lower(), set())
    for group in data.get("group") or []:
        result &= vtes.VTES.search["group"].get(group.lower(), set())
    for sect in data.get("sect") or []:
        result &= vtes.VTES.search["sect"].get(sect.lower(), set())
    for trait in data.get("trait") or []:
        result &= vtes.VTES.search["trait"].get(trait.lower(), set())
    for discipline in data.get("discipline") or []:
        discipline = config.DIS_MAP.get(discipline) or discipline
        result &= vtes.VTES.search["discipline"].get(discipline, set())
    for bonus in data.get("bonus") or []:
        result &= vtes.VTES.search.get(bonus.lower(), set())
    if data.get("text"):
        result &= set(vtes.VTES.search["text"].search(data["text"].lower()))
    return flask.jsonify(sorted(vtes.VTES.get_name(vtes.VTES[int(i)]) for i in result))


@base.route("/submit-ruling/<card>", methods=["POST"])
def submit_ruling(card):
    try:
        card = int(card)
    except ValueError:
        pass
    try:
        card = vtes.VTES.get_name(vtes.VTES[card])
    except KeyError:
        return "Card not found", 404
    data = flask.request.get_json() or {}
    text = data.get("text")
    link = data.get("link")
    if not (text and link):
        return "Invalid ruling data", 400
    if urllib.parse.urlparse(link).hostname not in {
        "boardgamegeek.com",
        "www.boardgamegeek.com",
        "groups.google.com",
        "www.vekn.net",
    }:
        return "Invalid ruling link", 400
    tryout = requests.get(link, stream=True)
    if not tryout.ok:
        return "Invalid ruling link", tryout.status_code

    url = "https://api.github.com/repos/lionel-panhaleux/krcg/issues"
    issue = {
        "title": card,
        "body": f"- **text:** {text}\n- **link:** {link}",
    }
    session = requests.session()
    session.auth = (os.getenv("GITHUB_USERNAME"), os.getenv("GITHUB_TOKEN"))
    response = session.post(url, json.dumps(issue))
    if response.ok:
        return flask.jsonify(response.json()), response.status_code
    else:
        return response.text, response.status_code



"""Static files generator for third parties

Produces static files for use in third parties softwares
"""

import argparse
import collections
import hashlib
import json
import logging
import os.path
import re
import sys

import requests

from . import twda
from . import vtes

logger = logging.getLogger()

parser = argparse.ArgumentParser(prog="krcg", description="VTES tool")


class FormatChoice(argparse.Action):
    # choices with nargs +/* : this is a known issue for argparse
    # cf. https://bugs.python.org/issue9625
    CHOICES = ["standard", "amaranth"]

    def __call__(self, parser, namespace, values, option_string=None):
        values = [v.lower() for v in values]
        if values:
            for value in values:
                if value not in self.CHOICES:
                    raise argparse.ArgumentError(
                        self, f"invalid choice: {value} (choose from {self.CHOICES})"
                    )
        setattr(namespace, self.dest, values)


parser.add_argument(
    "formats",
    metavar="FORMAT",
    nargs="+",
    action=FormatChoice,
    help="Generate static files in given formats",
)


def standard_rulings():
    result = {
        "rulings": {},
        "references": {},
        "cards": collections.defaultdict(list),
    }
    for key, data in vtes.VTES.items():
        if not isinstance(key, int):
            continue
        ruling_data = vtes.VTES.rulings.get(vtes.VTES.get_name(data))
        if not ruling_data:
            continue
        for ruling in ruling_data.get("Rulings", []):
            h = hashlib.sha3_256(ruling.encode("utf-8")).hexdigest()[-12:]
            result["rulings"].setdefault(h, ruling)
            result["cards"][key].append(h)
        for link in ruling_data.get("Rulings Links", []):
            result["references"].setdefault("[" + link["Reference"] + "]", link["URL"])
    return result


def amaranth_twda():
    try:
        amaranth_ids = {
            vtes.VTES.get_name(vtes.VTES[card["name"]]): card["id"]
            for card in (
                requests.get("http://amaranth.vtes.co.nz/api/cards").json()["result"]
            )
            if card["name"] in vtes.VTES
        }
    except requests.exceptions.ConnectionError as e:
        logger.exception("failed to connect to %s", e.request.url)
        exit(1)
    ret = []
    for twda_id, deck in twda.TWDA.items():
        try:
            ret.append(
                {
                    "author": deck.player,
                    "date": deck.date.format("YYYY-MM-DD"),
                    "title": deck.name or "",
                    "twda_key": twda_id,
                    "description": deck.event + ", " + deck.place,
                    "cards": {amaranth_ids[k]: v for k, v in deck.items()},
                }
            )
        except KeyError as e:
            logger.error(
                f"TWDA: {e.args[0]} not referenced in Amaranth, skipping {twda_id}"
            )
            continue
    return ret


def standard_twda():
    content = """<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN" "http://www.w3.org/TR/html4/loose.dtd">
<html><head><meta http-equiv="Content-Type" content="text/html; charset=utf-8">
<title>TWDA</title></head>

<body>
<center>
<h1>Tournament Winning Deck Archive</h1>
<h2>Formerly hosted on the Lasombra</h2>

<p>
This is a chronological archive of decks that have won tournaments that were<br>
sanctioned by the Vampire: Elder Kindred Network and had 10 or more players.<br>
National and Continental Championship are included as well, irrelatively to their attendance.<br>
To be included in this list, a tournament report had to be made on the official V:EKN Player's Forum:<br>
<a href="http://www.vekn.net/index.php/forum/9-event-reports-and-twd">http://www.vekn.net/index.php/forum/9-event-reports-and-twd</a><br>
This archive will not include the Storyline Tournament Winning Decks.<br>
Unless otherwise noted, each deck was the winning player's creation.<br>
<br>
The original <a href="http://www.thelasombra.com/hall_of_fame.htm">Hall of Fame</a> and the <a href="http://www.veknfrance.com/decks/twd.htm">Tournament Winning Deck Archive</a> were closed as of October 26, 2013.<br>
Many thanks to Jeff Thompson for maintaining them for all these years.
</p>
<table width="500" align="center" border="2" cellpadding="0">
<tbody><tr align="center">
<td colspan="10">You can jump to the end of a year that interests you:</td>
</tr>
<tr align="center">
<td colspan="4"></td>
<td><a href="#Year2020">2020</a></td>
<td><a href="#Year2019">2019</a></td>
<td><a href="#Year2018">2018</a></td>
<td><a href="#Year2017">2017</a></td>
<td colspan="3"></td>
</tr><tr align="center">
<td><a href="#Year2016">2016</a></td>
<td><a href="#Year2015">2015</a></td>
<td><a href="#Year2014">2014</a></td>
<td><a href="#Year2013">2013</a></td>
<td><a href="#Year2012">2012</a></td>
<td><a href="#Year2011">2011</a></td>
<td><a href="#Year2010">2010</a></td>
<td><a href="#Year2009">2009</a></td>
<td><a href="#Year2008">2008</a></td>
<td><a href="#Year2007">2007</a></td>
</tr><tr align="center">
<td><a href="#Year2006">2006</a></td>
<td><a href="#Year2005">2005</a></td>
<td><a href="#Year2004">2004</a></td>
<td><a href="#Year2003">2003</a></td>
<td><a href="#Year2002">2002</a></td>
<td><a href="#Year2001">2001</a></td>
<td><a href="#Year2000">2000</a></td>
<td><a href="#Year1999">1999</a></td>
<td><a href="#Year1998">1998</a></td>
<td><a href="#Year1997">1997</a></td>
</tr>
</tbody></table>
<p>Some abbreviations are used throughout the archive:
<table>
<tr><td>NC</td><td>National Championship</td><td>NAC</td><td>North American (Continental) Championship</td></tr>
<tr><td>NCQ</td><td>National Championship Qualifier</td><td>SAC</td><td>South American (Continental) Championship</td></tr>
<tr><td>ECQ</td><td>European Championship Qualifier</td><td>EC</td><td>European (Continental) Championship</td></tr>
<tr><td>CCQ</td><td>Continental Championship Qualifier</td><td>ACC</td><td>Asian Continental Championship</td></tr>
</table>"""  # noqa: E501
    decks = sorted(twda.TWDA.items(), key=lambda a: a[1].date, reverse=True)
    current_year = None
    for twda_id, deck in decks:
        if current_year != deck.date.year:
            current_year = deck.date.year
            content += f'\n<h3><a id="Year{current_year}">{current_year}</a></h3>\n\n'
        player = deck.player
        player = re.sub(r"\s*\([^\)]*\)", "", player)
        player = re.sub(r"\s*--\s+.*", "", player)
        if player[-1] != "s":
            player += "'s"
        else:
            player += "'"
        event = deck.event
        event = re.sub(r"\s*--\s+.*", "", event)
        place = deck.place
        place = re.sub(r"\s*,", ",", place)
        place = re.sub(r",,", ",", place)
        place = [x.strip() for x in place.split(",")[-2:]]
        event = event.strip().strip(":")
        if event[-len(place[0]) :] == place[0]:
            place = place[1:]
            event = event.strip().strip(":") + ","
        else:
            event = event.strip().strip(":") + ":"
        place = ", ".join(place)
        content += f"<a href=#{twda_id}>{player} {event} {place} {deck.date.strftime('%B %Y')}</a><br>\n"  # noqa: E501
    content += "</center>\n"
    for twda_id, deck in decks:
        content += f"<a id={twda_id} href=#>Top</a>\n<hr><pre>\n"
        content += vtes.VTES.deck_to_txt(deck, wrap=False)
        content += "\n</pre>\n"
    return content


def main():
    args = parser.parse_args(sys.argv[1:])
    logger.addHandler(logging.StreamHandler(sys.stderr))
    try:
        if not vtes.VTES:
            vtes.VTES.load_from_vekn(save=False)
        vtes.VTES.configure()
        if not twda.TWDA:
            twda.TWDA.load_from_vekn(save=False)
        twda.TWDA.configure(spoilers=False)
    except requests.exceptions.ConnectionError as e:
        logger.exception("failed to connect to %s", e.request.url)
        exit(1)
    out_folder = os.path.join(os.path.dirname(__file__), "..", "static")
    if "standard" in args.formats:
        with open(
            os.path.join(out_folder, "standard-rulings.json"), "w", encoding="utf-8",
        ) as out_file:
            json.dump(standard_rulings(), out_file, indent=2)
        with open(
            os.path.join(out_folder, "twd.htm"), "w", encoding="utf-8-sig",
        ) as out_file:
            out_file.write(standard_twda())
    if "amaranth" in args.formats:
        with open(
            os.path.join(out_folder, "amaranth-twda.json"), "w", encoding="utf-8",
        ) as out_file:
            json.dump(amaranth_twda(), out_file, indent=2)



"""Command-line interface.
"""
import argparse
import logging
import math
import sys

import arrow
import requests

from . import analyzer
from . import config
from . import twda
from . import vtes

logger = logging.getLogger()

# VTES configure must be done before parsing args
if vtes.VTES:
    vtes.VTES.configure()


def init(args):
    if args.cards:
        vtes.VTES.load_csv(args.file)
    elif args.twda:
        if not vtes.VTES:
            logger.critical(
                "VTES cards database must be initialized before TWDA database."
            )
            exit(1)
        vtes.VTES.configure()
        twda.TWDA.load_html(args.file)
    else:
        try:
            vtes.VTES.load_from_vekn()
            vtes.VTES.configure()
            twda.TWDA.load_from_vekn()
        except requests.exceptions.ConnectionError as e:
            logger.critical(
                "failed to connect to {}"
                " - check your Internet connection and firewall settings".format(
                    e.request.url
                )
            )
            exit(1)


def typical_copies(A, card_name):
    deviation = math.sqrt(A.variance[card_name])
    min_copies = max(1, round(A.average[card_name] - deviation))
    max_copies = round(A.average[card_name] + deviation)
    if min_copies == max_copies:
        ret = f"{min_copies}"
    else:
        ret = f"{min_copies}-{max_copies}"
    if max_copies > 1:
        ret += " copies"
    else:
        ret += " copy"
    return ret


def affinity(args):
    # build a condition matching options
    def condition(card):
        return (not args.crypt or vtes.VTES.is_crypt(card)) and (
            not args.library or vtes.VTES.is_library(card)
        )

    twda.TWDA.configure(args.date_from, args.date_to, args.players)
    A = analyzer.Analyzer()
    A.refresh(*args.cards, condition=condition, similarity=1)
    if len(A.examples) < 4:
        print("Too few example in TWDA.")
        if len(A.examples) > 0:
            print(
                "To see them:\n\tkrcg deck {}".format(
                    " ".join('"' + name + '"' for name in args.cards)
                )
            )
        return
    for name, score in A.candidates(*args.cards, no_filter=True)[: args.number or None]:
        score = round(score * 100 / len(args.cards))
        if args.minimum and args.minimum > score:
            break
        # do not include spoilers if affinity is within 50% of natural occurence
        if score < twda.TWDA.spoilers.get(name, 0) * 150:
            continue
        print(
            f"{name:<30} (in {score:.0f}% of decks, typically "
            f"{typical_copies(A, name)})"
        )


def _search(args):
    result = set(card["Id"] for card in vtes.VTES.original_cards.values())
    for type_ in args.type or []:
        result &= vtes.VTES.search["type"].get(type_.lower(), set())
    for clan in args.clan or []:
        clan = config.CLANS_AKA.get(clan.lower()) or clan
        result &= vtes.VTES.search["clan"].get(clan.lower(), set())
    for group in args.group or []:
        result &= vtes.VTES.search["group"].get(group.lower(), set())
    for bonus in args.bonus or []:
        result &= vtes.VTES.search.get(bonus.lower(), set())
    for trait in args.trait or []:
        result &= vtes.VTES.search["trait"].get(trait.lower(), set())
    for discipline in args.disc or []:
        discipline = config.DIS_MAP.get(discipline) or discipline
        result &= vtes.VTES.search["discipline"].get(discipline, set())
    for exclude in args.exclude_type or []:
        result -= vtes.VTES.search["type"].get(exclude.lower(), set())
    return result


def search(args):
    result = _search(args)
    for card_id in list(result)[: args.number]:
        card = vtes.VTES[int(card_id)]
        name = vtes.VTES.get_name(card)
        print(name)
        if args.full:
            print(_card_text(card))
            print()
    if len(result) > args.number:
        print(f"... ({len(result)} results available, use the -n option)")


def top(args):
    result = _search(args)
    twda.TWDA.configure(args.date_from, args.date_to, args.players)
    A = analyzer.Analyzer()
    A.refresh(condition=lambda c: vtes.VTES[c]["Id"] in result)
    for card_name, count in A.played.most_common()[: args.number]:
        card = vtes.VTES[card_name]
        print(
            f"{card_name:<30} (played in {count} decks, typically "
            f"{typical_copies(A, card_name)})"
        )
        if args.full:
            print(_card_text(card))
            print()


def build(args):
    twda.TWDA.configure(args.date_from, args.date_to, args.players)
    print(vtes.VTES.deck_to_txt(analyzer.Analyzer().build_deck(*args.cards)))


def deck_(args):
    twda.TWDA.configure(args.date_from, args.date_to, args.players, spoilers=False)
    decks = {i: twda.TWDA[i] for i in args.cards_or_id if i in twda.TWDA}
    cards = [
        vtes.VTES.get_name(vtes.VTES[c]) for c in args.cards_or_id if c not in twda.TWDA
    ]
    if not decks:
        A = analyzer.Analyzer()
        try:
            A.refresh(*cards, similarity=1)
        except analyzer.AnalysisError:
            exit(1)
        decks.update(A.examples)
    if len(decks) == 1:
        args.full = True
    if not args.full:
        print(f"-- {len(decks)} decks --")
    for twda_id, example in sorted(decks.items(), key=lambda a: a[1].date):
        if args.full:
            print(
                "[{:<15}]===================================================".format(
                    twda_id
                )
            )
            print(vtes.VTES.deck_to_txt(example))
        else:
            print("[{}] {}".format(twda_id, example))


def card(args):
    for i, name in enumerate(args.cards):
        if i > 0:
            print()
        try:
            name = int(name)
        except ValueError:
            pass
        try:
            card = vtes.VTES[name]
        except KeyError:
            logger.critical("Card not found: {}".format(name))
            exit(1)
        print(vtes.VTES.get_name(card))
        if args.short:
            continue
        print(_card_text(card))
        print(_card_rulings(args, vtes.VTES.get_name(card)))


def _card_text(card):
    text = "[{}]".format("/".join(card["Type"]))
    if card.get("Clan"):
        text += "[{}]".format("/".join(card["Clan"]))
    if card.get("Pool Cost"):
        text += "[{}P]".format(card["Pool Cost"])
    if card.get("Blood Cost"):
        text += "[{}B]".format(card["Blood Cost"])
    if card.get("Conviction Cost"):
        text += "[{}C]".format(card["Conviction Cost"])
    if card.get("Capacity"):
        text += "[{}]".format(card["Capacity"])
    if card.get("Group"):
        text += "(g.{})".format(card["Group"])
    if card.get("Burn Option"):
        text += "(Burn Option)"
    if card.get("Banned"):
        text += " -- BANNED in " + card["Banned"]
    text += " -- ({} - #{})".format(", ".join(card["Set"]), card["Id"])
    if "Disciplines" in card:
        text += "\n{}".format("/".join(card["Disciplines"]) or "-- No discipline")
    text += "\n{}".format(card["Card Text"])
    return text


def _card_rulings(args, card_name):
    rulings = vtes.VTES.rulings.get(card_name)
    if args.text or not rulings:
        return ""
    text = "\n-- Rulings\n"
    for ruling in rulings["Rulings"]:
        text += ruling + "\n"
    if args.links:
        for link in rulings["Rulings Links"]:
            text += f"[{link['Reference']}]: {link['URL']}\n"
    return text


def complete(args):
    for name in vtes.VTES.complete(args.name.lower()):
        print(name)


# ############################################################################# argparse
def add_deck_boundaries(parser):
    """Common arguments: --from and --to to control year boundaries of TWDA analysis.
    """
    parser.add_argument(
        "--from",
        type=lambda s: arrow.get(s, "YYYY"),
        default=arrow.get(2008, 1, 1),
        dest="date_from",
        help="do not consider decks that won before this year (default 2008-01-01)",
    )
    parser.add_argument(
        "--to",
        type=lambda s: arrow.get(s, "YYYY"),
        default=arrow.get(),
        dest="date_to",
        help="do not consider decks that won after this year",
    )
    parser.add_argument(
        "--players",
        type=int,
        default=0,
        help="do not consider decks with less players than this",
    )


class NARGS_CHOICE_WITH_ALIASES(argparse.Action):
    # choices with nargs +/* : this is a known issue for argparse
    # cf. https://bugs.python.org/issue9625
    CHOICES = []
    ALIASES = {}
    CASE_SENSITIVE = False

    def __call__(self, parser, namespace, values, option_string=None):
        if not self.CASE_SENSITIVE:
            values = [v.lower() for v in values]
        if values:
            for value in values:
                if value not in self.CHOICES and value not in self.ALIASES:
                    raise argparse.ArgumentError(
                        self, f"invalid choice: {value} (choose from {self.CHOICES})"
                    )
        setattr(
            namespace,
            self.dest,
            [
                value if value in self.CHOICES else self.ALIASES[value]
                for value in values
            ],
        )


class DisciplineChoice(NARGS_CHOICE_WITH_ALIASES):
    CHOICES = vtes.VTES.search.get("discipline", {}).keys()
    ALIASES = config.DIS_MAP
    CASE_SENSITIVE = True


class ClanChoice(NARGS_CHOICE_WITH_ALIASES):
    CHOICES = vtes.VTES.search.get("clan", {}).keys()
    ALIASES = config.CLANS_AKA


class TypeChoice(NARGS_CHOICE_WITH_ALIASES):
    CHOICES = vtes.VTES.search.get("type", {}).keys()


class TraitChoice(NARGS_CHOICE_WITH_ALIASES):
    CHOICES = vtes.VTES.search.get("trait", {}).keys()


class GroupChoice(NARGS_CHOICE_WITH_ALIASES):
    CHOICES = vtes.VTES.search.get("group", {}).keys()


class BonusChoice(NARGS_CHOICE_WITH_ALIASES):
    CHOICES = set(vtes.VTES.search.keys()) - {
        "trait",
        "type",
        "capacity",
        "group",
        "discipline",
        "clan",
        "text",
    }


def add_search(parser):
    """Common search arguments for top and search commands
    """
    parser.add_argument(
        "-n",
        "--number",
        type=int,
        default=10,
        help="Number of cards to print (default 10)",
    )
    parser.add_argument(
        "-d",
        "--discipline",
        action=DisciplineChoice,
        dest="disc",
        metavar="DISC",
        nargs="+",
        help="Filter by discipline ({})".format(
            ", ".join(vtes.VTES.search.get("discipline", {}).keys())
        ),
    )
    parser.add_argument(
        "-c",
        "--clan",
        action=ClanChoice,
        metavar="CLAN",
        nargs="+",
        help="Filter by clan ({})".format(
            ", ".join(vtes.VTES.search.get("clan", {}).keys())
        ),
    )
    parser.add_argument(
        "-t",
        "--type",
        action=TypeChoice,
        metavar="TYPE",
        nargs="+",
        help="Filter by type ({})".format(
            ", ".join(vtes.VTES.search.get("type", {}).keys())
        ),
    )
    parser.add_argument(
        "-g",
        "--group",
        action=GroupChoice,
        metavar="GROUP",
        nargs="+",
        help="Filter by group ({})".format(
            ", ".join(vtes.VTES.search.get("group", {}).keys())
        ),
    )
    parser.add_argument(
        "-e",
        "--exclude-type",
        action=TypeChoice,
        metavar="TYPE",
        nargs="+",
        help="Exclude given type ({})".format(
            ", ".join(vtes.VTES.search.get("type", {}).keys())
        ),
    )
    parser.add_argument(
        "-r",
        "--trait",
        action=TraitChoice,
        metavar="TRAIT",
        nargs="+",
        help="Filter by trait ({})".format(
            ", ".join(vtes.VTES.search.get("trait", {}).keys())
        ),
    )
    parser.add_argument(
        "-b",
        "--bonus",
        action=BonusChoice,
        metavar="BONUS",
        nargs="+",
        help="Filter by bonus ({})".format(", ".join(BonusChoice.CHOICES)),
    )
    parser.add_argument("-f", "--full", action="store_true", help="display card text")


root_parser = argparse.ArgumentParser(prog="krcg", description="VTES tool")
root_parser.add_argument(
    "-v",
    "--verbosity",
    type=int,
    default=0,
    metavar="N",
    help="0: errors, 1: warnings , 2: info, 3: debug",
)
subparsers = root_parser.add_subparsers(
    metavar="", title="subcommands", dest="subcommand"
)

# ################################################################################# init
parser = subparsers.add_parser("init", help="initialize the local TWDA database")
parser.add_argument(
    "-c", "--cards", action="store_true", help="Initialize VTES cards database"
)
parser.add_argument(
    "-t", "--twda", action="store_true", help="Initialize TWDA database"
)
parser.add_argument(
    "file",
    type=argparse.FileType("r", encoding="utf-8"),
    nargs="?",
    help="vtes.csv or TWDA.html file",
)
parser.set_defaults(func=init)

# ############################################################################# affinity
parser = subparsers.add_parser(
    "affinity", help="display cards with the most affinity to given cards"
)
add_deck_boundaries(parser)
parser.add_argument(
    "-n", "--number", type=int, default=0, metavar="N", help="Number of cards to print",
)
parser.add_argument(
    "-m",
    "--minimum",
    type=int,
    default=33,
    metavar="P",
    help="Minimum affinity to display (default 33)",
)
parser.add_argument("-c", "--crypt", action="store_true", help="Only crypt cards")
parser.add_argument("-l", "--library", action="store_true", help="Only library cards")
parser.add_argument(
    "cards",
    metavar="CARD",
    nargs="+",
    type=lambda a: vtes.VTES.get_name(vtes.VTES[a]),
    help="reference cards",
)
parser.set_defaults(func=affinity)

# ################################################################################## top
parser = subparsers.add_parser(
    "top", help="display top cards (played in most TW decks)"
)
add_deck_boundaries(parser)
add_search(parser)
parser.set_defaults(func=top)

# ################################################################################ build
parser = subparsers.add_parser("build", help="build a deck")
add_deck_boundaries(parser)
parser.add_argument(
    "cards",
    metavar="CARD",
    nargs="*",
    type=lambda a: vtes.VTES.get_name(vtes.VTES[a]),
    help="reference cards",
)
parser.set_defaults(func=build)

# ################################################################################# deck
parser = subparsers.add_parser("deck", help="show TWDA decks")
add_deck_boundaries(parser)
parser.add_argument(
    "-f", "--full", action="store_true", help="display each deck content"
)
parser.add_argument(
    "cards_or_id", metavar="TXT", nargs="*", help="list TWDA decks from ID or cards"
)
parser.set_defaults(func=deck_)

# ################################################################################# card
parser = subparsers.add_parser("card", help="show VTES cards")
parser.add_argument("-s", "--short", action="store_true", help="display only card name")
parser.add_argument(
    "-t", "--text", action="store_true", help="display card text only (no rulings)"
)
parser.add_argument("-l", "--links", action="store_true", help="display ruling links")
parser.add_argument("cards", metavar="CARD", nargs="+", help="card names or IDs")
parser.set_defaults(func=card)

# ############################################################################# complete
parser = subparsers.add_parser("complete", help="card name completion")
parser.add_argument("-f", "--full", action="store_true", help="display cards text")
parser.add_argument("name", metavar="NAME", help="partial name")
parser.set_defaults(func=complete)

# ############################################################################### search
parser = subparsers.add_parser("search", help="card search")
add_search(parser)
parser.set_defaults(func=search)


def main():
    args = root_parser.parse_args(sys.argv[1:])
    logger.addHandler(logging.StreamHandler(sys.stderr))
    logger.setLevel(
        {0: logging.ERROR, 1: logging.WARNING, 2: logging.INFO, 3: logging.DEBUG}[
            args.verbosity
        ]
    )
    if not args.subcommand:
        root_parser.print_help()
        return
    if args.subcommand != "init":
        if not vtes.VTES:
            logger.critical('VTES cards database is not initialized. Use "krcg init"')
            exit(1)
        if not twda.TWDA:
            logger.critical('TWDA database is not initialized. Use "krcg init"')
            exit(1)
    args.func(args)



"""Deck class: pickling and card access under conditions."""
import collections
import logging

import arrow
import arrow.parser

logger = logging.getLogger()


class Deck(collections.Counter):
    """A VTES deck, including meta information such as author and decription.

    Meta information:
        - name: Name of the deck
        - author: Deck author
        - comments: Comments on the deck
        - event: Event (for TWD)
        - place: Place where the event was held
        - date: Date on which the event was held
        - tournament_format: Format of the event
        - players_count: Count of players at the event
        - player: Player who played the deck
    """

    def __init__(self, *args, **kwargs):
        self.author = kwargs.pop("author", None)
        super().__init__(*args, **kwargs)
        self.event = None
        self.place = None
        self.date = None
        self.tournament_format = None
        self.players_count = 0
        self.player = None
        self.event_link = None
        self.score = None
        self.name = None
        self.cards_comments = {}
        self.comments = ""

        # for comments parsing
        self.separator = False
        self.comment = None
        self.card_name = None
        self.line_num = None
        self.multiline = False
        self.marker = False

    def comment_begin(
        self, line_num, comment, card_name=None, multiline=False, marker=False
    ):
        if self.comment:
            self._assign_comment()
        self.line_num = line_num
        if not self.separator or marker:
            self.comment = comment.strip(" ")
        else:
            self.comment = comment.strip(" ()[]-/*")
        self.card_name = card_name
        self.multiline = multiline
        self.marker = marker

    def maybe_comment_line(self, line_num, comment, card_name=None):
        if self.comment:
            if self.separator:
                comment = comment.strip(" ()[]-/*")
            self.comment += "\n" + comment
        else:
            self.comment_begin(
                line_num,
                comment,
                card_name or self.card_name,
                not self.separator,
                not self.separator,
            )

    def comment_end(self):
        if self.comment:
            self._assign_comment()
        self.comment = None
        self.card_name = None
        self.line_num = None
        self.multiline = False
        self.marker = False

    def _assign_comment(self):
        self.comment = self.comment.strip("\n")
        if not self.comment:
            return
        log_version = self.comment.replace("\n", " ").strip()
        multiline = len(self.comment.split("\n")) > 1
        if not self.marker:
            if not multiline:
                logger.warning(f"[{self.line_num}] failed to parse [{log_version}]")
                return
            logger.debug(f"[{self.line_num}] unmarked comment [{log_version}]")
        if multiline and not self.multiline:
            logger.debug(
                f"[{self.line_num}] unexpected multiline comment [{log_version}]"
            )
        comment = self.comment + "\n"
        if self.card_name:
            self.cards_comments[self.card_name] = comment
            return
        if self.comments:
            self.comments += "\n"
        self.comments += comment

    def __getstate__(self):
        """For pickle serialization.

        Deck inherits `dict` and its special handling of pickle.
        """
        return {
            "cards": collections.OrderedDict(self.cards()),
            "author": self.author,
            "event": self.event,
            "place": self.place,
            "date": self.date.format("MMMM Do YYYY") if self.date else None,
            "tournament_format": self.tournament_format,
            "score": self.score,
            "players_count": self.players_count,
            "player": self.player,
            "event_link": self.event_link,
            "name": self.name,
            "cards_comments": self.cards_comments,
            "comments": self.comments,
        }

    def __setstate__(self, state):
        """For pickle deserialization.
        """
        self.author = state.get("author")
        self.event = state.get("event")
        self.place = state.get("place")
        try:
            self.date = arrow.get(state.get("date"), "MMMM Do YYYY")
        except arrow.parser.ParserError:
            pass
        self.tournament_format = state.get("tournament_format")
        self.score = state.get("score")
        self.players_count = int(state.get("players_count", 0))
        self.player = state.get("player")
        self.event_link = state.get("event_link")
        self.name = state.get("name")
        self.comments = state.get("comments", "")
        self.cards_comments = state.get("cards_comments", {})
        self.update(state.get("cards", {}))

    def __reduce__(self):
        """For pickle serialization."""
        return (Deck, (), self.__getstate__())

    def __str__(self):
        return self.name or "(No Name)"

    def cards(self, condition=None):
        """Generator yielding (card_name, count), with an optional filter.

        Args:
            condition (func): (opt.) Condition each card must validate to be selected

        Yields:
            card, count (str, int)
        """
        for card, count in self.items():
            if condition and not condition(card):
                continue
            yield card, count

    def card_names(self, condition=None):
        """Generator yielding card names with an optional filter.

        Args:
            condition (func): (opt.) Condition each card must validate to be selected

        Yields:
            card (str)
        """
        for card, _count in self.cards(condition):
            yield card

    def cards_count(self, condition=None):
        """Card counts with an optional filter.

        Args:
            condition (func): (opt.) Condition each card must validate to be selected

        Returns:
            int: the count of all cards (matching the condition if any)
        """
        return sum(count for _card, count in self.cards(condition))
